import datetime
import json
import re
import functools
import typing


def _unhandled_exceptions_return(return_value: bool) -> typing.Callable:
    """
    Create a decorator that converts all unhandled exceptions to the given
    return value.
    """

    def decorator(func: typing.Callable) -> typing.Callable:
        """Decorate the given input function."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """
            Convert all non keyboard interupt exceptions to the given
            value
            """
            try:
                return func(*args, **kwargs)
            except KeyboardInterrupt as err:
                raise err
            except Exception as err:
                return return_value
        return wrapper

    return decorator


class JsonStringChecker():
    """Checker for validating that a string is json serializable."""

    name = 'json'

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value can be deserialized to json."""
        json.loads(value)
        return True


class ChoiceChecker():
    """Checker for validating that a value is in a set of choices."""

    name = 'choice'

    def __init__(self, options: list = None):
        self._options = set(options)

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value is a valid choice"""
        return value in self._options


class IsoTimestampChecker():
    """Checker for validating that a value is an iso timestamp."""

    name = 'iso_timestamp'

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value is a valid iso timestamp."""
        datetime.datetime.fromisoformat(value)
        return True


class RegexChecker():
    """Checker for validating that a value meets the given regex."""

    name = 'regex'

    def __init__(self, regex: str = None, ignore_case=False):
        flags = re.IGNORECASE if ignore_case else 0
        self._regex = re.compile(regex, flags)

    @_unhandled_exceptions_return(False)
    def __call__(self, value: typing.Any) -> bool:
        """Confirm the given value matches the regex."""
        return bool(self._regex.match(value))


class EmailChecker(RegexChecker):
    """Checker for validating that a value is a valid email."""

    name = 'email'

    def __init__(self):
        super().__init__(
            regex=r'^[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,63}$',
            ignore_case=True
        )


class UuidChecker(RegexChecker):
    """Checker for validating that a value is a valid email."""

    name = 'uuid'

    def __init__(self):
        pat = r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'
        super().__init__(regex=pat, ignore_case=True)


CHECKERS = [
    ChoiceChecker,
    EmailChecker,
    IsoTimestampChecker,
    JsonStringChecker,
    RegexChecker,
    UuidChecker,
]
CHECKER_MAP = {c.name: c for c in CHECKERS}



import json

import pytest

from horkos.cmdline import check

CONTENT_TYPES = [
    'foo\nbar\nbaz',
    json.dumps([{'foo': 'bar'}, {'foo': 'baz'}]),
    json.dumps({'foo': 'bar'}) + '\n' + json.dumps({'foo': 'baz'}),
]


@pytest.mark.parametrize('content', CONTENT_TYPES)
def test_convert_to_records_csv(content):
    """Should be able to convert several file types to a list of records."""
    result = check._convert_to_records(content)

    assert result[0]['foo'] == 'bar'
    assert result[1]['foo'] == 'baz'



import json
from unittest import mock
import uuid

import pytest

from horkos import checkeromatic


def test_json_string_checker_happy_path():
    """Should be able check a valid json string."""
    value = json.dumps({'foo': 'bar'})
    checker = checkeromatic.JsonStringChecker()

    passes = checker(value)

    assert passes


def test_json_string_checker_bad_json():
    """Should be able check an invalid json string."""
    checker = checkeromatic.JsonStringChecker()

    passes = checker('<xml></xml>')

    assert not passes


def test_json_string_checker_wrong_type():
    """Should be able check a value of the wrong type."""
    checker = checkeromatic.JsonStringChecker()

    passes = checker(12)

    assert not passes


def test_choice_checker_happy_path():
    """
    Should be able to check that a value is within a valid set of
    choices.
    """
    checker = checkeromatic.ChoiceChecker(options=['foo', 'bar'])

    passes = checker('foo')

    assert passes


def test_choice_checker_bad_value():
    """
    Should be able to see that a value is not within a valid set of
    choices.
    """
    checker = checkeromatic.ChoiceChecker(options=['foo', 'bar'])

    passes = checker('fizbuzz')

    assert not passes


def test_email_checker_happy_path():
    """Should be able to validate an email."""
    checker = checkeromatic.EmailChecker()

    passes = checker('could.be@an-email.com')

    assert passes


def test_email_checker_bad_value():
    """Should be able to identify an invalid email."""
    checker = checkeromatic.EmailChecker()

    passes = checker('for-sure-not-valid')

    assert not passes


def test_uuid_checker_happy_path():
    """Should be able to validate a uuid."""
    value = str(uuid.uuid4())
    checker = checkeromatic.UuidChecker()

    passes = checker(value)

    assert passes


def test_uuid_checker_bad_value():
    """Should be able to identify a non-uuid."""
    value = '398l159z-8m4n-455o-pqr0-75s66tu074jk'
    checker = checkeromatic.UuidChecker()

    passes = checker(value)

    assert not passes


def test_iso_timestamp_checker_happy_path():
    """Should be able to identify a iso compliant timestamp."""
    value = '2020-03-18T12:34:56'
    checker = checkeromatic.IsoTimestampChecker()

    passes = checker(value)

    assert passes


@mock.patch('json.loads')
def test_keyboard_interupts_surface(loads: mock.MagicMock):
    """A keyboard interupt should always get reraised."""
    loads.side_effect = KeyboardInterrupt
    checker = checkeromatic.JsonStringChecker()

    with pytest.raises(KeyboardInterrupt):
        checker('some-value')



import typing

from horkos import definitions
from horkos import errors


def _int(value: typing.Any) -> int:
    """Convert the given value into an integer."""
    if isinstance(value, float) and (int(value) - value) != 0:
        raise ValueError('Cannot losslessly convert to an integer')
    return int(value)


def _bool(value: typing.Any) -> bool:
    """Convert the given value to a boolean"""
    if isinstance(value, bool):
        return value
    if isinstance(value, str) and value.lower() == 'true':
        return True
    if isinstance(value, str) and value.lower() == 'false':
        return False
    if isinstance(value, int) and value == 0:
        return False
    if isinstance(value, int) and value == 1:
        return True
    raise ValueError(f'Cannot convert {value} to a boolean')


def cast(record: dict, field_map: typing.Dict[str, definitions.Field]) -> dict:
    """
    Cast the fields on the given record to the expected types.

    :param record:
        The record on which to check for incorrect types. It is expected to be
        a dictionary mapping field names to values.
    :param field_map:
        A dictionary mapping field names to their field objects.
    :return:
        A dictionary representing the record with each value converted to
        the expected type.
    """
    cast_map = {
        'boolean': _bool,
        'float': float,
        'integer': _int,
        'string': str,
    }

    errors_set = []
    result = {}
    for field_name, field in field_map.items():
        if field_name not in record and field.required:
            errors_set.append(f'{field_name} is required')
            continue
        value = record.get(field_name)
        if value is None and not field.nullable:
            errors_set.append(f'{field_name} cannot be null')
            continue
        if value is None:
            result[field_name] = value
            continue
        try:
            result[field_name] = cast_map[field.type_](value)
        except (TypeError, ValueError):
            errors_set.append(
                f'value of "{value}" for {field_name} '
                f'could not be cast to {field.type_}'
            )
    if errors_set:
        msg = f'Casting errors - {", ".join(errors_set)}'
        raise errors.RecordValidationError(msg)
    return result



############################
#
# WITHIN DATASET
#
############################
#
# Fields
# - Naming
# -

# Data sets


# Field naming
# Data set naming
# Names should mean the same thing
# Type consistency
# Documentation length



import typing


class Field(typing.NamedTuple):
    """The schema of a specific field in the dataset."""

    name: str
    type_: str
    description: str
    required: bool
    nullable: bool
    checks: typing.Dict[str, typing.Callable]
    labels: list



import functools
import inspect
import io
import typing
import yaml

from horkos import casters
from horkos import checkeromatic
from horkos import definitions
from horkos import errors

Schemaable = typing.Union[str, dict, 'Schema', io.TextIOWrapper]


class Schema(typing.NamedTuple):
    """The schema of the whole dataset."""

    name: str
    description: str
    labels: dict
    fields: typing.Dict[str, definitions.Field]

    def process(self, record: dict) -> dict:
        """
        Process a record, validating that it meets all the schema's checks.
        If it does not meet the requirements an exception is raised.

        :param record:
            The record to process.
        :return:
            The processed record with values cast and validated.
        """
        error_set = []
        casted = casters.cast(record, self.fields)
        for field_name, field in self.fields.items():
            value = casted[field_name]
            if value is None and field.nullable:
                continue
            for check_name, check in field.checks:
                if not check(value):
                    error_set.append(
                        f'value of "{value}" in {field_name} did not '
                        f'pass {check_name} check'
                    )
        if error_set:
            raise errors.RecordValidationError(
                f'Check errors - {", ".join(error_set)}'
            )
        return casted


def _schema_from_dict(
        raw_schema: dict,
        strict: bool = True,
        custom_checkers: typing.Dict[str, typing.Callable] = None
) -> Schema:
    """
    Create a schema from a raw schema dictionary.

    :param raw_schema:
        A raw schema matching the expected json/yaml schema format.
    :param strict:
        Whether to operate in strict mode. If True descriptions will be
        required.
    :param custom_checkers:
        A dictionary of custom checkers. This dictionary should map the name
        of the check to a function that generates the check function
    :return:
        The resulting schema.
    """
    standardized = _standardize_schema(raw_schema, strict=strict)
    checkers = checkeromatic.CHECKER_MAP.copy()
    checkers.update(custom_checkers or {})
    fields = {}
    raw_fields = standardized['fields']
    for field_name, field_body in raw_fields.items():
        checks = []
        for i, raw_check in enumerate(field_body['checks']):
            position = f'fields.{field_name}.checks[{i}]'
            check_name = raw_check['name']
            if check_name not in checkers:
                raise errors.SchemaValidationError(
                    f'In "{position}" check name "{check_name}" is unknown.'
                )
            check = _create_check_func(
                checkers[check_name],
                raw_check['args'],
                position
            )
            checks.append((check_name, check))
        fields[field_name] = definitions.Field(
            field_name,
            field_body['type'],
            field_body['description'],
            field_body['required'],
            field_body['nullable'],
            checks,
            field_body['labels'],
        )
    return Schema(
        standardized['name'],
        standardized['description'],
        standardized['labels'],
        fields,
    )

def _load_yaml(
        filepath_or_handle: typing.Union[str, io.TextIOWrapper]
) -> dict:
    """
    Read the given yaml file.

    :param filepath_or_handle:
        The path to the yaml file to convert into a schema.
    :return:
        The resulting schema.
    """
    if not isinstance(filepath_or_handle, str):
        return yaml.safe_load(filepath_or_handle)
    with open(filepath_or_handle) as f:
        return yaml.safe_load(f)


def _create_check_func(
        check: typing.Callable, args: dict, position: str
) -> typing.Callable:
    """
    Create the actual check function, baking the args into it. The returned
    function should only expect a single positional argument to check.

    :param check:
        The base check. This is either a class that will be initialized
        with the given arguments, or a function that will be partially filled
        with the given args.
    :param args:
        The arguments to initialize the check with.
    :param position:
        The position of the check within the schema. This is used to provide
        more clear error messages.
    :return:
        A callable that expects a single positional argument that it will
        check.
    """
    try:
        if inspect.isclass(check):
            return check(**args)
        signature = inspect.signature(check)
        signature.bind('value', **args)
    except TypeError:
        raise errors.SchemaValidationError(
            f'Invalid arguments in "{position}"'
        )
    return functools.partial(check, **args)


def _standardize_schema(raw_schema: dict, strict: bool = True) -> dict:
    """
    Standardize the given raw_schema with all optional fields filled out
    with defaults. If the raw schema is not valid a SchemaValidationError will
    be raised.

    :param raw_schema:
        The schema to standardize.
    :param strict:
        Whether to operate in strict mode. If True descriptions will be
        required.
    :return:
        The standardized schema.
    """
    standardized = {}
    top = [
        ('name', str, True),
        ('description', str, strict),
        ('labels', dict, False),
        ('fields', dict, True),
    ]
    for key, type_, required in top:
        if key not in raw_schema and required:
            raise errors.SchemaValidationError(f'"{key}" is required')
        value = raw_schema[key] if key in raw_schema else type_()
        if not isinstance(value, type_):
            raise errors.SchemaValidationError(
                f'"{key}" must be of type {type_.__name__}'
            )
        standardized[key] = value
    standardized['fields'] = {
        field_id: _standardize_field(raw_field, field_id, strict=strict)
        for field_id, raw_field in raw_schema['fields'].items()
    }
    return standardized


def _standardize_field(
        raw_field: dict, field_id: str, strict: bool = True
) -> dict:
    """
    Standardize a given raw field with all the optional parts filled out
    with defaults. If the raw field can't be validated a SchemaValidationError
    will be raised.

    :param raw_field:
        The raw field to standardize.
    :param field_id:
        The id of the field being standardized.
    :param strict:
        Whether to operate in strict mode. If True descriptions will be
        required.
    :return:
        The standardized field.
    """
    standardized = {}
    top = [
        ('type', str, True, None),
        ('description', str, strict, str),
        ('labels', dict, False, dict),
        ('checks', list, False, list),
        ('required', bool, False, lambda: True),
        ('nullable', bool, False, bool),
    ]
    for key, type_, required, default in top:
        working = f'fields.{field_id}.{key}'
        if key not in raw_field and required:
            raise errors.SchemaValidationError(f'"{working}" is required')
        value = raw_field[key] if key in raw_field else default()
        if not isinstance(value, type_):
            raise errors.SchemaValidationError(
                f'"{working}" must be of type {type_.__name__}'
            )
        standardized[key] = value
    standardized['checks'] = [
        _standardize_check(raw_check, f'fields.{field_id}.check[{i}]')
        for i, raw_check in enumerate(raw_field.get('checks') or [])
    ]
    return standardized


def _standardize_check(
        raw_check: typing.Union[str, dict], prefix: str
) -> dict:
    """
    Convert a check into the standard format. If the check cannot be
    standardize a schema validation error is raised.

    :param raw_check:
        The raw_check to standardize.
    :param prefix:
        The document position prefix to integrate into error messages.
    :return:
        A standard check dictionary.
    """
    if isinstance(raw_check, str):
        return {'name': raw_check, 'args': {}}
    if isinstance(raw_check, dict):
        if 'name' not in raw_check:
            raise errors.SchemaValidationError(
                f'"{prefix}.name" is required'
            )
        if not isinstance(raw_check['name'], str):
            raise errors.SchemaValidationError(
                f'"{prefix}.name" must be a string'
            )
        if 'args' in raw_check and not isinstance(raw_check['args'], dict):
            raise errors.SchemaValidationError(
                f'"{prefix}.args" must be a dictionary'
            )
        return {'name': raw_check['name'], 'args': raw_check.get('args', {})}
    raise errors.SchemaValidationError(
        f'"{prefix}" must be a string or dictionary'
    )


def load_schema(
        schema: Schemaable,
        strict: bool = False,
        custom_checkers: typing.Dict[str, typing.Callable] = None
) -> Schema:
    """
    Load a schema in from a dictionary, filepath, or copy it from an existing
    schema.

    :param raw_schema:
        A raw schema matching the expected json/yaml schema format.
    :param strict:
        Whether to operate in strict mode. If True descriptions will be
        required.
    :param custom_checkers:
        A dictionary of custom checkers. This dictionary should map the name
        of the check to a function that generates the check function
    """
    working = schema
    custom_checkers = custom_checkers or {}
    if isinstance(working, Schema):
        return working
    if not isinstance(working, dict):
        working = _load_yaml(working)
    return _schema_from_dict(
        working, strict=strict, custom_checkers=custom_checkers
    )



import pytest

from horkos import casters
from horkos import definitions
from horkos import errors


def test_cast_happy_path():
    """Should be able to cast each of a records values to its expected type."""
    bool_field = definitions.Field(
        name='bool_field',
        type_='boolean',
        description='Some description',
        required=True,
        nullable=False,
        checks={},
        labels={},
    )
    field_map = {
        'bool_field_1': bool_field._replace(name='bool_field_1'),
        'bool_field_2': bool_field._replace(name='bool_field_2'),
        'bool_field_3': bool_field._replace(name='bool_field_3'),
        'bool_field_4': bool_field._replace(name='bool_field_4'),
        'bool_field_5': bool_field._replace(name='bool_field_5'),
        'int_field': definitions.Field(
            name='int_field',
            type_='integer',
            description='Some description',
            required=True,
            nullable=False,
            checks={},
            labels={},
        ),
        'str_field': definitions.Field(
            name='str_field',
            type_='string',
            description='Some description',
            required=True,
            nullable=False,
            checks={},
            labels={},
        ),
    }
    record = {
        'bool_field_1': 'true',
        'bool_field_2': 'False',
        'bool_field_3': 1,
        'bool_field_4': 0,
        'bool_field_5': True,
        'int_field': '123',
        'str_field': 12345
    }

    cast = casters.cast(record, field_map)

    assert cast['bool_field_1'] is True
    assert cast['bool_field_2'] is False
    assert cast['bool_field_3'] is True
    assert cast['bool_field_4'] is False
    assert cast['bool_field_5'] is True
    assert cast['int_field'] == 123
    assert cast['str_field'] == '12345'


def test_cast_raises_all_errors():
    """Should raise all errors when present."""
    field_map = {
        'bool_field': definitions.Field(
            name='bool_field',
            type_='boolean',
            description='Some description',
            required=True,
            nullable=False,
            checks={},
            labels={},
        ),
        'float_field': definitions.Field(
            name='float_field',
            type_='float',
            description='Some description',
            required=True,
            nullable=False,
            checks={},
            labels={},
        ),
        'int_field_1': definitions.Field(
            name='int_field_1',
            type_='integer',
            description='Some description',
            required=True,
            nullable=False,
            checks={},
            labels={},
        ),
        'int_field_2': definitions.Field(
            name='int_field_2',
            type_='integer',
            description='Some description',
            required=True,
            nullable=False,
            checks={},
            labels={},
        ),
        'str_field_1': definitions.Field(
            name='str_field_1',
            type_='string',
            description='Some description',
            required=True,
            nullable=False,
            checks={},
            labels={},
        ),
        'str_field_2': definitions.Field(
            name='str_field_2',
            type_='string',
            description='Some description',
            required=True,
            nullable=False,
            checks={},
            labels={},
        ),
    }
    record = {
        'bool_field': 'T',
        'float_field': 'word',
        'int_field_1': 1.123,
        'int_field_2': 'word',
        'str_field_1': None,
    }

    with pytest.raises(errors.RecordValidationError) as err:
        casters.cast(record, field_map)

    msg = 'value of "T" for bool_field could not be cast to boolean'
    assert msg in str(err.value)
    msg = 'value of "word" for float_field could not be cast to float'
    assert msg in str(err.value)
    msg = 'value of "1.123" for int_field_1 could not be cast to integer'
    assert msg in str(err.value)
    msg = 'value of "word" for int_field_2 could not be cast to integer'
    assert msg in str(err.value)
    msg = 'str_field_1 cannot be null'
    assert msg in str(err.value)
    msg = 'str_field_2 is required'
    assert msg in str(err.value)



import os
import typing

import pytest

from horkos import errors
from horkos import schemaomatic

DIR_PATH = os.path.dirname(os.path.realpath(__file__))
RESOURCE_PATH = os.path.join(DIR_PATH, 'resources')


def test_load_schema():
    """Should be able to create a schema from a yaml file without error."""
    yaml_path = os.path.join(RESOURCE_PATH, 'sample.yaml')

    schemaomatic.load_schema(yaml_path)


def test_schema_process():
    """Should be able to process a record with the schema."""
    raw_schema = {
        'name': 'foobar',
        'description': 'The details',
        'fields': {
            'first_field': {
                'type': 'string',
                'description': 'This is a description',
                'checks': [
                    {'name': 'regex', 'args': {'regex': r'\d+'}}
                ]
            },
            'second_field': {
                'type': 'string',
                'description': 'Some description',
                'required': False,
                'nullable': True,
                'checks': [
                    'json'
                ]
            }
        }
    }
    schema = schemaomatic.load_schema(raw_schema)

    result = schema.process({'first_field': 123456})

    assert result['first_field'] == '123456'
    assert result['second_field'] == None


def test_schema_process_with_errors():
    """Should get all check errors at once."""
    raw_schema = {
        'name': 'foobar',
        'description': 'The details',
        'fields': {
            'first_field': {
                'type': 'string',
                'description': 'This is a description',
                'checks': [
                    {'name': 'regex', 'args': {'regex': r'\d+'}}
                ]
            },
            'second_field': {
                'type': 'string',
                'description': 'Some description',
                'nullable': True,
                'checks': [
                    'json'
                ]
            }
        }
    }
    schema = schemaomatic.load_schema(raw_schema)

    with pytest.raises(errors.RecordValidationError) as err:
        schema.process({'first_field': 'foobar', 'second_field': 'fizz'})

    err_msg = 'value of "foobar" in first_field did not pass regex check'
    assert err_msg in str(err.value)
    err_msg = 'value of "fizz" in second_field did not pass json check'
    assert err_msg in str(err.value)


SCENARIOS = [
    ({}, '"prefix.name" is required'),
    ({'name': []}, '"prefix.name" must be a string'),
    ({'name': 'something', 'args': []}, '"prefix.args" must be a dictionary'),
    ([], '"prefix" must be a string or dictionary'),
]


@pytest.mark.parametrize('check, msg', SCENARIOS)
def test_standardize_check_errors(check: typing.Any, msg: str):
    """If given a bad check an error should be raised."""
    with pytest.raises(errors.SchemaValidationError) as err:
        schemaomatic._standardize_check(check, 'prefix')

    assert msg in str(err.value)


SCENARIOS = [
    ({'type': 'string'}, '"fields.field_id.description" is required'),
    (
        {'type': 'bool', 'description': 'foo', 'labels': []},
        '"fields.field_id.labels" must be of type dict'
    ),
]


@pytest.mark.parametrize('field, msg', SCENARIOS)
def test_standardize_field_errors(field: typing.Any, msg: str):
    """If given a bad field an error should be raised."""
    with pytest.raises(errors.SchemaValidationError) as err:
        schemaomatic._standardize_field(field, 'field_id')

    assert msg in str(err.value)


SCENARIOS = [
    ({}, '"name" is required'),
    ({'name': {}}, '"name" must be of type str'),
]


@pytest.mark.parametrize('schema, msg', SCENARIOS)
def test_standardize_schema_errors(schema: typing.Any, msg: str):
    """If given a bad schema an error should be raised."""
    with pytest.raises(errors.SchemaValidationError) as err:
        schemaomatic._standardize_schema(schema)

    assert msg in str(err.value)


SCENARIOS = [
    {'name': 'custom', 'args': {'foo': 'bar', 'fizz': 'buzz'}},
    {'name': 'custom', 'args': {}},
    {'name': 'regex', 'args': {'any': 'thing'}},
]


@pytest.mark.parametrize('check', SCENARIOS)
def test_schema_with_bad_checker(check: dict):
    """If given a bad checker a validation error should be raised."""

    def custom(value, foo, boo='faz'):
        """Custom check function"""

    raw_schema = {
        'name': 'name',
        'fields': {
            'fieldId': {'type': 'str', 'checks': [check]}
        }
    }

    with pytest.raises(errors.SchemaValidationError) as err:
        schemaomatic.load_schema(
            raw_schema, strict=False, custom_checkers={'custom': custom}
        )

    assert 'Invalid arguments in "fields.fieldId.checks[0]"' in str(err.value)


def test_schema_with_unknown_checker():
    """A SchemaValidationError should be raised if we don't know the checker"""
    raw_schema = {
        'name': 'name',
        'fields': {
            'fieldId': {'type': 'str', 'checks': ['unknown']}
        }
    }

    with pytest.raises(errors.SchemaValidationError) as err:
        schemaomatic.load_schema(raw_schema, strict=False)

    msg = 'In "fields.fieldId.checks[0]" check name "unknown" is unknown.'
    assert msg in str(err.value)


def test_load_schema_with_custom_checker():
    """A custom checker should be handled without error."""
    raw_schema = {
        'name': 'name',
        'fields': {
            'fieldId': {'type': 'str', 'checks': ['custom']}
        }
    }
    custom = {'custom': lambda x: True}

    schemaomatic.load_schema(
        raw_schema, custom_checkers=custom
    )



import pytest

from horkos import cataloger


def test_catalog_process_happy_path():
    """Should be able to process a record through a catalog."""
    schema = {
        'name': 'dataset',
        'fields': {'foo': {'type': 'string'}}
    }
    catalog = cataloger.Catalog([schema], strict=False)

    record = catalog.process('dataset', {'foo': 'bar'})

    assert record['foo'] == 'bar'


def test_catalog_process_sad_path():
    """Should be able to process a record through a catalog."""
    catalog = cataloger.Catalog()

    with pytest.raises(ValueError) as err:
        catalog.process('doesnt-exist', {'foo': 'bar'})

    assert 'No schema exists for "doesnt-exist"' in str(err.value)


def test_catalog_update():
    """Should be able to update an existing catalog."""
    catalog = cataloger.Catalog(strict=False)
    schema = {
        'name': 'dataset',
        'fields': {'foo': {'type': 'string'}}
    }
    catalog.update(schema)

    record = catalog.process('dataset', {'foo': 'bar'})

    assert record['foo'] == 'bar'



import io
import os
from unittest import mock

from horkos import cmdline

DIR_PATH = os.path.dirname(os.path.realpath(__file__))
RESOURCE_PATH = os.path.join(DIR_PATH, 'resources')


def test_check_cmdline():
    """Check cmdline should be able to run."""
    schema_file = os.path.join(RESOURCE_PATH, 'sample.yaml')
    data_file = os.path.join(RESOURCE_PATH, 'sample.csv')
    output = io.StringIO()

    with mock.patch('sys.stdout', new=output):
        cmdline.main(['check', '--schema', schema_file, data_file])

    result = output.getvalue()
    assert 'value of "ERROR" in method did not pass choice check' in result
    expected = 'value of "BAD" for response_code could not be cast to integer'
    assert expected in result
    assert '2 errors found' in result



"""Check that the data matches the declared schema."""
import argparse
import csv
import io
import json
import sys
import typing

from horkos import errors
from horkos import schemaomatic


def configure_parser(parser: argparse.ArgumentParser):
    """Configure the given parser to work for the check command."""
    parser.add_argument(
        '--schema', '-s',
        dest='schema_file',
        required=True,
        type=argparse.FileType('r'),
        help='The schema file describing the data.'
    )
    parser.add_argument(
        'datafile',
        nargs='?',
        type=argparse.FileType('r'),
        default=sys.stdin,
    )


def _convert_to_records(contents: str) -> typing.List[dict]:
    """
    Convert a file's contents into a set of records. The contents are expected
    to be either a csv with a header of column names, a json list, or a
    set of json records seperated by a new line.

    :param contents:
        The contents of the file to parse.
    :return:
        A list of records from the file.
    """
    try:
        rows = json.loads(contents)
        if isinstance(rows, list):
            return rows
    except ValueError:
        pass
    try:
        rows = [json.loads(r) for r in contents.strip().split('\n')]
        if all(isinstance(r, dict) for r in rows):
            return rows
    except ValueError:
        pass
    reader = csv.DictReader(io.StringIO(contents))
    return [
        {k: v if v != '' else None for k, v in r.items()}
        for r in reader
    ]


def main(
        datafile: io.TextIOWrapper,
        schema_file: io.TextIOWrapper,
):
    """Validate that the input file matches the given schema."""
    schema = schemaomatic.load_schema(schema_file)
    contents = datafile.read()
    records = _convert_to_records(contents)
    error_count = 0
    for i, record in enumerate(records):
        try:
            schema.process(record)
        except errors.RecordValidationError as err:
            error_count += 1
            print(f'[Row {i}]: {err}', file=sys.stdout)
    msg = f'{error_count} error{"s" if error_count != 1 else ""} found'
    print(msg, file=sys.stdout)



import io
import typing

from horkos import schemaomatic


class Catalog:
    """A collection of schemas."""

    def __init__(
            self,
            schemas: typing.List[schemaomatic.Schemaable] = None,
            strict: bool = True,
    ):
        """
        Initialize the catalog of data schemas.

        :param schema_files:
            Yaml file paths of schema files to include in the catalog.
        :param schemas:
            Existing schema objects to include in the catalog.
        :param strict:
            Whether to operate in strict mode. If True descriptions will be
            required.
        """
        self.strict = strict
        schemas = schemas or []
        all_schemas = [
            schemaomatic.load_schema(s, strict=strict) for s in schemas
        ]
        self.schemas = {s.name: s for s in all_schemas}

    def process(
            self,
            name: str,
            record: dict,
    ) -> dict:
        """
        Process a record, validating that it meets all the schema's checks.
        If it does not meet the requirements an exception is raised.

        :param name:
            The name of the data being processed. This is used to identify
            the schema to check the record against.
        :param record:
            The record to process.
        :return:
            The processed record with values cast and validated.
        """
        schema = self.schemas.get(name)
        if schema is None:
            raise ValueError(f'No schema exists for "{name}"')
        return schema.process(record)

    def update(self, schema: schemaomatic.Schemaable):
        """
        Update the catalog with the given schema.

        :param schema:
            The schema to add/update to the catalog.
        """
        to_add = schemaomatic.load_schema(schema, strict=self.strict)
        self.schemas[to_add.name] = to_add



class Error(Exception):
    """Base error."""


class SchemaValidationError(Error):
    """The given data schema failed to validate."""


class RecordValidationError(Error):
    """A record failed to validate against the schema."""



import argparse

from horkos.cmdline import check


def _parse(args: list = None) -> dict:
    """Parse the given arguments and return them as a simple dictionary."""
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='subcommand', required=True)
    check_parser = subparsers.add_parser('check', help=check.__doc__)
    check.configure_parser(check_parser)
    return vars(parser.parse_args(args))


def main(args: list = None):
    """Run the horkos command."""
    parsed_args = _parse(args)
    handlers = {
        'check': check.main,
    }
    command = parsed_args.pop('subcommand')
    handlers[command](**parsed_args)



import setuptools

setuptools.setup(
    name='horkos',
    version='0.0.4',
    description='A package for validating incoming data.',

    author='Kevin Schiroo',
    author_email='kjschiroo@gmail.com',
    license='MIT',
    url='https://gitlab.com/kjschiroo/horkos',

    packages=setuptools.find_packages(),
    entry_points = {
        'console_scripts': ['horkos=horkos.cmdline:main'],
    },
    install_requires=['pyyaml'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.7',
)
